#
# Copyright (c) 2000 Going Virtual, LLC. All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Contact Information:
#
# Going Virtual, LLC
# 10800 Independence Pointe Pkwy
# Matthews, NC 28105
# www.goingv.com
#
# Robert E. Tierney 
# bob.tierney@goingv.com 
# 1 (704) 849-0731 x109
#
# 
# Change History
#
# 07/31/2000 - Fixed typo in _prepareStatement.  Thanks to Brad Clements.
#
#
__doc__ = """ Going Virtual Interbase Database API

Python Database API Specification 2.0 Compliant

$Id: gvibCursor.py,v 1.7 2000/07/31 14:32:02 btierney Exp $
"""
__version__ = '$Revision: 1.7 $'[11:-2]

import types
import string
import gvibBase
import gvibTypes
import gvibExceptions

from gvibMessages import messageTable


class Cursor:
    """
    Cursor Class - Instances of this class represent a database cursor.

    This class represents a database cursor and allows users to execute SQL
    statements against an open database connection.  The user instantiates
    a Cursor object like so:

    import gvib
    connection = gvib.connect('database', 'user', 'password')
    cursor     = connection.cursor()
    
    """

    arraysize = 1

    __INVALID_STATE = -1
    __CREATED_STATE =  0
    __OPEN_STATE    =  1
    __CLOSED_STATE  =  2

    def __init__(self, py_connect, ib_connect, name, cols):
        self.__pyConnect   = py_connect
        self.__connect     = ib_connect
	self.__cursor      = gvibBase.IBCursor(ib_connect, cols)
	self.__sql_stmt    = ''
	self.__stmt_type   = -1
	self.__state       = Cursor.__CREATED_STATE

	self.cursor_name   = name
	self.rowcount      = -1
	self.description   = None
	self.sqlcounts     = { 'select' : -1,
	                       'insert' : -1, 
			       'update' : -1, 
			       'delete' : -1 }


    def setinputsizes(self, sizes):
        pass


    def setoutputsizes(self, size, column = -1):
        pass


    def execute(self, sql_stmt, parms = ()):
        """
	execute - Execute an arbitrary SQL statement.

        This method prepares a SQL statement, binds output columns, binds input
	parameters and executes the SQL statement.  The parameters are
	iterated over and bound to each parameter marker in the statement. 
	This has implications for a singleton string parameter.  The parameters
	are indexed using a 'for n in range(len(parms)):' mechanism.  If you
	have a single parameter marker and it is a string you must pass it as
	as singleton tuple or a single item list. However, if you have one
	or more parameter markers that are single characters, you pass a bare
        string which contains a character for each marker. For example:

	cr.execute('select dchar from dtable where dchar in (?,?,?)', 'ABC')
	cr.fetchall()

	[('A',), ('B',), ('C',)]

	"""

        # Ensure cursor wasn't closed by user.
	self._checkStatus()

	self.__first_time = 1

	# Make sure the parms are of the right type. 
	from types import StringType, TupleType, ListType
	if type(parms) not in (StringType, TupleType, ListType):
	    msg = messageTable['badParmType']
	    raise gvibExceptions.Error(ibmsg = msg);

        cursor  = self.__cursor
	connect = self.__connect

	# If a cursor is open then close it but retain the statement context
	# for reuse.
	self._closeRetaining()

	# Reset the rowcount and sqlcounts attributes
	self.rowcount = -1
	for n in ['select', 'insert', 'update', 'delete']:
	    self.sqlcounts[n] = -1

	# If the SQL statement to be executed is different than the previous
	# one then prepare it.
	if self.__sql_stmt != sql_stmt:
	    self._prepareStatement(sql_stmt)
	    self._getStatementType()

	    if self.description != None:
	        self.description = None

	    # If data will be returned then get the description.
	    if cursor.osqld > 0:
	        self._getColDesc()


	# Make sure the user provided enough parameters to satisfy the SQL
	# statement.
	if len(parms) < cursor.isqld:
	    msg = messageTable['noParms']
	    raise gvibExceptions.ProgrammingError(ibmsg = msg)


        # Loop through the parameters and bind each one in the input SQLDA.
        # The user must provide them in the correct order of course.  We
	# don't use the return value.  It's used for exception purposes.
	import gvib
	for pm in range(len(parms)):
	    data = parms[pm]
	    if (isinstance(data, gvib.Date) or 
	        isinstance(data, gvib.Time) or
		isinstance(data, gvib.Timestamp)):
	      data = parms[pm].tuple()

	    rc = gvibBase.IBCursorBindParm(cursor, pm, data)


	# Submit the SQL to Interbase for execution.
	if self.__stmt_type == gvibTypes.isc_info_sql_stmt_exec_procedure:
            rc = gvibBase.isc_dsql_execute2(cursor.status,
	              connect.tr_handle,
	              cursor.stmt_handle,
		      1,
		      cursor.isqlda,
		      cursor.osqlda);
	else:
            rc = gvibBase.isc_dsql_execute(cursor.status,
	              connect.tr_handle,
	              cursor.stmt_handle,
		      1,
		      cursor.isqlda);

        if rc: gvibBase.IBSetException(cursor.status)

        # If the SQL just executed was DDL then we will commit it now.
	# Otherwise if there is output data and the user supplied a name
	# then set the cursor name.
	if self.__stmt_type == gvibTypes.isc_info_sql_stmt_ddl:
	    self.__pyConnect.commit()
	else:
	    if (self.cursor_name != None) and (cursor.osqld > 0):
	        rc = gvibBase.isc_set_cursor_name(
	                      cursor.status,
		              cursor.stmt_handle,
			      self.cursor_name,
			      0);

                if rc: gvibBase.IBSetException(cursor.status)

        # If the statement doesn't return any results then assume it's not
	# a select-type statement and thus get the rows-affected value now.
	if cursor.osqld == 0:
            self._getRowCount()
        else:
	    self.__state = Cursor.__OPEN_STATE



    def executemany(self, sql_stmt, parms = ()):
        """
        executemany - Execute the same SQL statement multiple times.

        This method determines how many times to execute the SQL statement
        by using the 'for n in parms:' syntax.  The SQL to be exceute must
        not return any results.
        """

        # Ensure cursor wasn't closed by user.
        self._checkStatus()

        for pm in parms:
            self.execute(sql_stmt, pm)
	    self._closeRetaining(0)                # 0 = disregard close errors



    def fetchone(self):
        """
	fetchone - Fetch one row from the result set.

	The resulting data are returned to the caller as a tuple.
	"""

	# Ensure cursor wasn't closed by user.
	self._checkStatus()

        cursor = self.__cursor

	# Does the latest execute return any data?
	if cursor.osqld == 0:
	    msg = messageTable['noReturnData']
	    raise gvibExceptions.ProgrammingError(ibmsg = msg)

        rc = gvibBase.isc_dsql_fetch(cursor.status,
	              cursor.stmt_handle,
		      1,
		      cursor.osqlda)

        if rc == 0L: 
	    result = gvibBase.IBCursorGetRow(cursor)
	elif rc == 100L:
            result = None
	    self._closeRetaining()
        else:
	    gvibBase.IBSetException(cursor.status)

        if self.__first_time:
            self._getRowCount()
	    self.__first_time = 0

        return result


        
    def fetchmany(self, size = None):
        """
	fetchmany - Fetch an arbitrary number of rows from the result set.

	The user may pass a number for how many rows to fetch.  If not 
	specified the arraysize attribute will be used as the number to
	fetch.  The resulting data are returned to the caller as a list 
	of tuples.
	"""

	# Ensure cursor wasn't closed by user.
	self._checkStatus()

	if size == None:
	    size = Cursor.arraysize

	if size < 1:
	    msg = messageTable['badRowCount']
	    raise gvibExceptions.Error(ibmsg = msg)

        rowList = []
	rowCount = size
	while rowCount > 0:
	#for n in range(size):
	    tmp = self.fetchone()
	    if tmp == None: 
	        break
            rowList.append(tmp)

	    rowCount = rowCount - 1

        return rowList



    def fetchall(self):
        """
	fetchall - Fetch all of any remaining rows from the result set.

	The data are returned to the caller as a list of tuples.
	"""

	# Ensure cursor wasn't closed by user.
	self._checkStatus()

        rowList = []
	tmp = self.fetchone()
	while tmp != None:
	    rowList.append(tmp)
	    tmp = self.fetchone()

        return rowList



    def callproc(self, name, parms = ()):
	"""
	callproc - Execute an executable-type stored procedure.

	This method may only be used to execute a stored procedure using
	the EXECUTE PROCEDURE syntax.  When using a stored procedure as the 
	data source in a select statement the execute and fetchXXX methods 
	should be used.  This method accepts positional parameters and 
	returns at most 1 row of data.  The data are returned as a tuple.
	"""

	# Ensure cursor wasn't closed by user.
	self._checkStatus()

        self.execute(name, parms)

        cursor = self.__cursor

	result = gvibBase.IBCursorGetRow(cursor)

	self.rowcount = len(result) > 0

	self._closeRetaining(0)                 # 0 = disregard close errors

	return result



    def close(self):
        """
	close - Close a cursor and render it unusable.
	
	After this method has been called the user MUST instantiate a new
	cursor using the connection object.
	"""

	# Ensure cursor wasn't closed by user.
	self._checkStatus()

        cursor  = self.__cursor
	connect = self.__connect

        if self.__state == Cursor.__OPEN_STATE:
            rc = gvibBase.isc_dsql_free_statement(cursor.status,
	                  cursor.stmt_handle,
	                  gvibTypes.DSQL_drop)

            if rc: gvibBase.IBSetException(cursor.status)

        self.__state = Cursor.__INVALID_STATE

        del self.__cursor



    def _closeRetaining(self, raise_error = 1):
        """
	_closeRetainging - Close a cursor but retain the context for reuse.
	"""

        cursor  = self.__cursor
	connect = self.__connect

        if self.__state == Cursor.__OPEN_STATE:
            rc = gvibBase.isc_dsql_free_statement(cursor.status,
	                  cursor.stmt_handle,
	                  gvibTypes.DSQL_close)

            if rc and raise_error: gvibBase.IBSetException(cursor.status)

            self.__state = Cursor.__CLOSED_STATE



    def _checkStatus(self):
        """"
	_checkStatus - Check to see if the cursor is still usable.
	"""

        if self.__state == Cursor.__INVALID_STATE:
	    msg = messageTable['badCursor']
	    raise gvibExceptions.Error(ibmsg = msg)



    def _getNewStmtHandle(self):
	"""
	_getNewStmtHandle - Drop and allocate a new statement handle.
	"""

        cursor  = self.__cursor
	connect = self.__connect

	rc = gvibBase.isc_dsql_free_statement(cursor.status,
	              cursor.stmt_handle,
		      gvibTypes.DSQL_drop);

        cursor.stmt_handle = 0L;

	rc = gvibBase.isc_dsql_allocate_statement(cursor.status,
	              connect.db_handle,
		      cursor.stmt_handle);

        if rc: gvibBase.IBSetException(cursor.status)



    def _prepareStatement(self, new_sql_stmt):
        """
	_prepareStatement - Prepare and describe an SQL statement.
	"""

        cursor  = self.__cursor
	connect = self.__connect

	# Prepare the SQL statement.
        #
	self.__sql_stmt = new_sql_stmt
	rc = gvibBase.isc_dsql_prepare(cursor.status,
	              connect.tr_handle,
                      cursor.stmt_handle,
                      0,
                      self.__sql_stmt,
                      gvibTypes.SQL_DIALECT_CURRENT,
                      cursor.osqlda)

        if rc: gvibBase.IBSetException(cursor.status)


	# If more XSQLVAR structures are needed then reallocate to get more.
	# We don't need to test the rc from the reallocate because it is 
	# used by Python to raise an exception, if one is thrown by the
	# reallocate function.
        #
        if cursor.osqln < cursor.osqld:
            rc = gvibBase.IBCursorReallocXDA(cursor, 'osqlda')
            rc = gvibBase.isc_dsql_describe(cursor.status,
	                  cursor.stmt_handle,
		          1,
		          cursor.osqlda)

	    if rc: gvibBase.IBSetException(cursor.status)


        # Bind all of the output columns, i.e. allocate appropriate space
	# for each column in the output XSQLVAR structures.  Again, rc
	# is only used for exception purposes.
        #
	rc = gvibBase.IBCursorBindCols(cursor);


	# Determine how many parameters are required by the SQL statement.
        #
        rc = gvibBase.isc_dsql_describe_bind(cursor.status,
	              cursor.stmt_handle,
	              1,
                      cursor.isqlda)

        if rc: gvibBase.IBSetException(cursor.status)


	# If more XSQLVAR structures are needed then reallocate to get more.
	# We don't need to test the rc from the reallocate because it is 
	# used by Python to raise an exception, if one is thrown by the
	# reallocate function.
        # 07/31/2000 - Fix typo IBCursoReallocXDA to IBCursorReallocXDA

        if cursor.isqln < cursor.isqld:
            rc = gvibBase.IBCursorReallocXDA(cursor, 'isqlda')
            rc = gvibBase.isc_dsql_describe_bind(cursor.status,
	                  cursor.stmt_handle,
	                  1,
	                  cursor.isqlda)

            if rc: gvibBase.IBSetException(cursor.status)



    def _getStatementType(self):
        """
	_getStatementType - Determine the type of SQL being executed.

	See section SQL Information Return Values in gvibTypes.py 
	for a list of recognized SQL statements.
	"""

        cursor = self.__cursor
        buffer = '\0x00' * 64

        rc = gvibBase.isc_dsql_sql_info(
	              cursor.status,
		      cursor.stmt_handle,
		      1,
		      chr(gvibTypes.isc_info_sql_stmt_type),
		      len(buffer),
		      buffer);

        if rc: gvibBase.IBSetException(cursor.status)

	if ord(buffer[0]) == gvibTypes.isc_info_sql_stmt_type:
	    length = gvibBase.isc_vax_integer(buffer[1:], 2)
	    self.__stmt_type = gvibBase.isc_vax_integer(buffer[3:], length)
	else:
	    self.__stmt_type = -1



    def _getRowCount(self):
        """
	_getRowCount - Get the number of records affected by the SQL statement.
	"""

	for n in ['select', 'insert', 'update', 'delete']:
	    self.sqlcounts[n] = -1

	cursor = self.__cursor
	buffer = '\0x00' * 64 

	rc = gvibBase.isc_dsql_sql_info(
	              cursor.status,
		      cursor.stmt_handle,
		      1,
		      chr(gvibTypes.isc_info_sql_records),
		      len(buffer),
		      buffer);

        if rc: gvibBase.IBSetException(cursor.status)

	if ord(buffer[0]) != gvibTypes.isc_info_sql_records:
            self.rowcount = -1
	    return;

	n = 3
	for k in range(4):
	    type   = ord(buffer[n])
	    length = gvibBase.isc_vax_integer(buffer[n + 1:], 2)
	    count  = gvibBase.isc_vax_integer(buffer[n + 3:], length)

            if type == gvibTypes.isc_info_req_select_count:
	        self.sqlcounts['select'] = count
	    elif type == gvibTypes.isc_info_req_insert_count:
	        self.sqlcounts['insert'] = count
	    elif type == gvibTypes.isc_info_req_update_count:
	        self.sqlcounts['update'] = count
	    elif type == gvibTypes.isc_info_req_delete_count:
	        self.sqlcounts['delete'] = count
	    else:
	        pass

	    n = n + 3 + length


	if (self.__stmt_type == gvibTypes.isc_info_sql_stmt_select or
	    self.__stmt_type == gvibTypes.isc_info_sql_stmt_select_for_upd):
	    self.rowcount = self.sqlcounts['select']
	elif self.__stmt_type == gvibTypes.isc_info_sql_stmt_insert:
	    self.rowcount = self.sqlcounts['insert']
	elif self.__stmt_type == gvibTypes.isc_info_sql_stmt_update:
	    self.rowcount = self.sqlcounts['update']
	elif self.__stmt_type == gvibTypes.isc_info_sql_stmt_delete:
	    self.rowcount = self.sqlcounts['delete']
	else:
	    self.rowcount = -1


    def _getColDesc(self):
	"""
	_getColDesc - Set the description attribute that describes the result
	              set columns.
	"""
	cursor = self.__cursor

        col_desc = gvibBase.IBCursorGetColDesc(cursor)

        # colDesc contains a list of tuples that describes the each column
	# of the result set.  However, the data is not as informative as
	# it could be.  So we will tweak it to make it so.

        T = gvibTypes

	self.description = []
	for row in col_desc:
	    name, alias, type, dsize, isize, prec, scale, nulls, subtype = row

	    # See if we really have a NUMERIC or DECIMAL column.

	    if type in (T.SQL_SHORT, T.SQL_LONG, T.SQL_INT64):
	        if subtype == 0:
		    pass
	        elif subtype == 1:
		    type = T.SQL_NUMERIC
                elif subtype == 2:
		    type = T.SQL_DECIMAL
                else:
		    pass

	    # See if rdb$db_key (row id) is one of the columns selected.
	    # This is the best kludge I could come up with because to
	    # Interbase the rdb$db_key columns is a CHAR column. To make
	    # matters worse, Interbase returns DB_KEY as the column name.
	    # So let the user beware.

	    lc_name  = string.lower(name)
	    lc_alias = string.lower(alias)

	    if lc_name == 'db_key' or lc_alias == 'db_key':
	        type = T.SQL_ROWID

            new_row = (alias, type, dsize, isize, prec, scale, nulls)

	    self.description.append(new_row)

	    
